from config import AIIAConfig
from torch import nn
import torch


class AIIA(nn.Module):
    def __init__(self, config: AIIAConfig):
        super(AIIA, self).__init__()
        self.config = config

    def save(self, model_path, config_path):
        torch.save(self.state_dict(), model_path)
        self.config.save(config_path)

    @classmethod
    def load(cls, config_path, model_path):
        config = AIIAConfig.load(config_path)
        model = cls(config)
        model.load_state_dict(torch.load(model_path))
        return model

class AIIABase(AIIA):
    def __init__(self, config: AIIAConfig):
        super(AIIABase, self).__init__(config)
        layers = []
        in_channels = config.num_channels
        for _ in range(config.num_hidden_layers):
            layers.extend([
                nn.Conv2d(in_channels, config.hidden_size, kernel_size=config.kernel_size, padding=1),
                getattr(nn, config.activation_function)(),
                nn.MaxPool2d(kernel_size=2)
            ])
            in_channels = config.hidden_size
        self.cnn = nn.Sequential(*layers)

    def forward(self, x):
        return self.cnn(x)

class AIIAExpert(AIIA):
    def __init__(self, config: AIIAConfig):
        super(AIIAExpert, self).__init__(config)
        self.base_cnn = AIIABase(config)

    def forward(self, x):
        return self.base_cnn(x)

class AIIAmoe(AIIA):
    def __init__(self, config: AIIAConfig, num_experts: int = 3):
        super(AIIAmoe, self).__init__(config)
        self.experts = nn.ModuleList([AIIAExpert(config) for _ in range(num_experts)])
        self.gate = nn.Sequential(
            nn.Linear(config.hidden_size, num_experts),
            nn.Softmax(dim=1)
        )

    def forward(self, x):
        expert_outputs = torch.stack([expert(x) for expert in self.experts], dim=1)
        gate_weights = self.gate(torch.mean(expert_outputs, (2, 3)))
        merged_output = torch.sum(expert_outputs * gate_weights.unsqueeze(2).unsqueeze(3), dim=1)
        return merged_output

class AIIAchunked(AIIA):
    def __init__(self, config: AIIAConfig, patch_size: int = 16):
        super(AIIAchunked, self).__init__(config)
        self.patch_size = patch_size
        self.base_cnn = AIIABase(config)

    def forward(self, x):
        patches = x.unfold(2, self.patch_size, self.patch_size).unfold(3, self.patch_size, self.patch_size)
        patches = patches.contiguous().view(patches.size(0), patches.size(1), -1, self.patch_size, self.patch_size)
        patch_outputs = []
        for p in torch.split(patches, 1, dim=2):
            p = p.squeeze(2)
            po = self.base_cnn(p)
            patch_outputs.append(po)
        combined_output = torch.mean(torch.stack(patch_outputs, dim=0), dim=0)
        return combined_output

class AIIAresursive(AIIA):
    def __init__(self, config: AIIAConfig, recursion_depth: int = 2):
        super(AIIAresursive, self).__init__(config)
        self.recursion_depth = recursion_depth
        self.chunked_cnn = AIIAchunked(config)

    def forward(self, x, depth=0):
        if depth == self.recursion_depth:
            return self.chunked_cnn(x)
        else:
            patches = x.unfold(2, 16, 16).unfold(3, 16, 16)
            patches = patches.contiguous().view(patches.size(0), patches.size(1), -1, 16, 16)
            processed_patches = []
            for p in torch.split(patches, 1, dim=2):
                p = p.squeeze(2)
                pp = self.forward(p, depth + 1)
                processed_patches.append(pp)
            combined_output = torch.mean(torch.stack(processed_patches, dim=0), dim=0)
            return combined_output
